package javademo;

import java.awt.Container;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.AdjustmentEvent;
import java.awt.event.AdjustmentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.swing.SwingUtilities;

public class MainFrame extends javax.swing.JFrame implements MouseMotionListener, MouseListener, MouseWheelListener, KeyListener {
  private int           peerlink_hit_size    = 12;
  public  final Object  mutex                = new Object();
  private final Point2D move_start           = new Point2D.Double();
  private       boolean first_render         = true;
  private       boolean render_everything    = false;
  private       boolean move_viewarea        = false;
  public final Map<Long,CustomIcon> icons    = Collections.synchronizedMap(new LinkedHashMap<>());
  public final List<CustomIcon> icons_zorder = Collections.synchronizedList(new LinkedList<>());
  public final List<CustomIcon> icons_selected = Collections.synchronizedList(new LinkedList<>());

  
////////////////////////////////////////////////////////////////////////////////
//Listeners Events Start
////////////////////////////////////////////////////////////////////////////////
  @Override
  public void mouseWheelMoved(MouseWheelEvent evt) {
    Rectangle visible = networkpanel.getVisibleRect();
    Double d = evt.getPreciseWheelRotation();
    if(!CustomIsKeyPressed.isKeyPressed(KeyEvent.VK_SHIFT)) {
      if(d <= 0) {
        visible.y -= 100;
      } else {
        visible.y += 100;
      }
    } else {
      if(d <= 0) {
        visible.x -= 100;
      } else {
        visible.x += 100;
      }
    }
    networkpanel.scrollRectToVisible(visible);
  }


  @Override
  public void mouseDragged(MouseEvent evt) {
    boolean needs_repaint = false;

    if(SwingUtilities.isLeftMouseButton(evt)) {
      synchronized(mutex) {
        if(!icons_selected.isEmpty()) {
          Iterator<CustomIcon> i = icons_selected.iterator();
          i = icons_selected.iterator();
          while(i.hasNext()) {
            CustomIcon ri = i.next();
            if(ri.icon_state_test(CustomIcon.SELECTED)) {
              Point new_loc = new Point((int)(evt.getX() - ri.icon_offset.getX()),(int)(evt.getY() - ri.icon_offset.getY()));
              new_loc.x = (new_loc.x / 10) * 10;
              new_loc.y = (new_loc.y / 10) * 10;
              ri.icon_set_location(new_loc);
              needs_repaint = true;
            }
          }
        }
      }
    }

    if(needs_repaint) {
      networkpanel.repaint();
    }
  }

  @Override
  public void mousePressed(MouseEvent evt) {
    networkpanel.requestFocus();

    if(SwingUtilities.isLeftMouseButton(evt)) {
      CustomIcon ri = radio_get_clicked(evt.getPoint());

      _radio_unselect_all();
      icons_selected.clear();

      Point mp = evt.getPoint();

      if(ri != null) {
        synchronized(mutex) {
          ri.icon_offset.setLocation(mp.getX() - ri.icon_get_x(), mp.getY() - ri.icon_get_y());
          _radio_set_zorder(ri, 0);
          if(ri.icon_state_test(CustomIcon.SELECTED)) {
            ri.icon_state(CustomIcon.SELECTED,0);
          } else {
            ri.icon_state(0,CustomIcon.SELECTED);
          }

          if(ri.icon_state_test(CustomIcon.SELECTED)) {
            if(!icons_selected.contains(ri)) {
              icons_selected.add(ri);
            }
          } else {
            icons_selected.remove(ri);
          }

        }
      }

      synchronized(mutex) {
        if(icons_selected.size() > 0) {
           Iterator<CustomIcon> i = icons_selected.iterator();
           while(i.hasNext()) {
             CustomIcon tri = i.next();
             tri.icon_offset.setLocation(evt.getX() - tri.icon_get_x(), evt.getY() - tri.icon_get_y());
           }
         }
      }
    }

    networkpanel.repaint();
  }

  @Override
  public void mouseReleased(MouseEvent evt) {
    synchronized(mutex) {
      _networkpanel_boundery(icons_selected);
    }
    networkpanel.repaint();
  }

  private void network_paint(Graphics g) {
    Graphics2D g2d = (Graphics2D)g;
    Rectangle visible = networkpanel.getVisibleRect();
    //FIXME:
    //g2d.setColor(ePeerLink.icon_default_color());
    //g2d.setStroke(ePeerLink.icon_default_stroke());

    if(!render_everything) {
      g2d.clip(visible);
    }

    synchronized(mutex) {
      for(int i = icons_zorder.size() - 1;i >= 0;i--) {
        CustomIcon ri = icons_zorder.get(i);
        Rectangle2D rradio = ri.icon_get_bounds();
        //TODO: Fix Me Maybe render all, every time but only render when needed
        //something like double buffer but more like static buffer unless graphic change has happend
        if(render_everything || rradio.intersects(visible)) {
          ri.icon_paint(g);
        }

      }
    }
  }

  private final CustomPainter networkpanel_painter = new CustomPainter() {
    @Override
    public void paint(Graphics g) {
      Graphics2D g2d = (Graphics2D) g;
      RenderingHints rh = new RenderingHints(null);
      rh.put(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
      rh.put(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
      rh.put(RenderingHints.KEY_DITHERING,  RenderingHints.VALUE_DITHER_ENABLE);
      rh.put(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
      rh.put(RenderingHints.KEY_STROKE_CONTROL, RenderingHints.VALUE_STROKE_NORMALIZE);
      rh.put(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
      rh.put(RenderingHints.KEY_FRACTIONALMETRICS, RenderingHints.VALUE_FRACTIONALMETRICS_ON);
      g2d.setRenderingHints(rh);

      network_paint(g);
    }
  };
////////////////////////////////////////////////////////////////////////////////
//Listeners Events Stop
////////////////////////////////////////////////////////////////////////////////

  public MainFrame() {
    initComponents();

    scr_network.getVerticalScrollBar().setUnitIncrement(100);
    scr_network.getHorizontalScrollBar().setUnitIncrement(100);
    networkpanel.addMouseMotionListener(this);
    networkpanel.addMouseListener(this);
    networkpanel.addMouseWheelListener(this);
    networkpanel.addKeyListener(this);
    scr_network.enableInputMethods(false);


    ((CustomPanel)networkpanel).setPainter(networkpanel_painter,false);

    scr_network.getHorizontalScrollBar().addAdjustmentListener(new AdjustmentListener() {
      @Override
      public void adjustmentValueChanged(AdjustmentEvent e) {
        networkpanel.repaint();
      }
    });

    scr_network.getVerticalScrollBar().addAdjustmentListener(new AdjustmentListener() {
      @Override
      public void adjustmentValueChanged(AdjustmentEvent e) {
        networkpanel.repaint();
      }
    });
    
    for(int i = 0; i < 1089; i++) {
      CustomIcon ci = new CustomIcon(this);
      ci.icon_set_icon("res/images/icon_radio_small.png");
      icons.put(Long.valueOf(i), ci);
      icons_zorder.add(0, ci);
      radio_auto_place(ci,icons.size() - 1);
    }
    
    center_scroll();
  }

  private void center_scroll() {
    Rectangle visible = networkpanel.getVisibleRect();
    Rectangle bounds = networkpanel.getBounds();
    visible.y = (int)(((bounds.getHeight() / 2) - (visible.getHeight() / 2)));
    visible.x = (int)(((bounds.getWidth() / 2) - (visible.getWidth() / 2)));
    networkpanel.scrollRectToVisible(visible);
  }

  //////////////////////////////////////////////////////////////////////////////
  //  CustomPanel Stuff
  //////////////////////////////////////////////////////////////////////////////
  private void _networkpanel_boundery(Collection<CustomIcon> c) {
    Iterator<CustomIcon> i = c.iterator();
    int cnt = 0;
    int w =  (int)(networkpanel.getWidth());
    int h =  (int)(networkpanel.getHeight());
    boolean got_stuck = false;
    while(i.hasNext()) {
      CustomIcon ri = i.next();
      int x_edge = (int)((ri.icon_get_x() + ri.icon_get_width())) - 1;
      int y_edge = (int)((ri.icon_get_y() + ri.icon_get_height())) - 1;
      int x = (int)(ri.icon_get_x());
      int y = (int)(ri.icon_get_y());
      int x_offset = 0;
      int y_offset = 0;
      if(x < 0 || y < 0 || x_edge > w || y_edge > h) {
        if(x < 0) {
          x_offset = Math.abs(x);
        } else if(x_edge > w) {
          x_offset = Math.abs(x_edge - w) * -1;
        }
        if(y < 0) {
          y_offset = Math.abs(y);
        } else if(y_edge > h) {
          y_offset = Math.abs(y_edge - h) * -1;
        }
        Iterator<CustomIcon> j = c.iterator();
        while(j.hasNext()) {
          CustomIcon tri = j.next();
          tri.icon_set_x(tri.icon_get_x() + x_offset);
          tri.icon_set_y(tri.icon_get_y() + y_offset);
        }
        i = c.iterator();
      }
      if(cnt == 10000) {
        got_stuck = true;
        break;
      }
      cnt++;
    }
    if(got_stuck) {
      System.out.println("Fix: Got Stuck");
      i = c.iterator();
      while(i.hasNext()) {
        CustomIcon ri = i.next();
        int x_edge = (int)((ri.icon_get_x() + ri.icon_get_width())) - 1;
        int y_edge = (int)((ri.icon_get_y() + ri.icon_get_height())) - 1;
        int x = (int)(ri.icon_get_x());
        int y = (int)(ri.icon_get_y());
        int x_offset = 0;
        int y_offset = 0;
        if(x < 0 || y < 0 || x_edge > w || y_edge > h) {
          if(x < 0) {
            x_offset = Math.abs(ri.icon_get_x());
          } else if(x_edge > w) {
            x_offset = Math.abs(x_edge - w) * -1;
          }
          if(y < 0) {
            y_offset = Math.abs(ri.icon_get_y());
          } else if(y_edge > h) {
            y_offset = Math.abs(y_edge - h) * -1;
          }
          ri.icon_set_x(ri.icon_get_x() + x_offset);
          ri.icon_set_y(ri.icon_get_y() + y_offset);
        }
      }
    }
  }

  private void _radio_unselect_all_except(CustomIcon rii) {
    Iterator<CustomIcon> i = icons.values().iterator();
    while(i.hasNext()) {
      CustomIcon ri = i.next();
      if(rii == null || !ri.equals(rii)) {
        ri.icon_state(CustomIcon.SELECTED, 0);
      }
    }
  }

  private void _radio_unselect_all() {
    _radio_unselect_all_except(null);
  }

  int origin_x = 0;
  int origin_y = 0;
  public void radio_auto_place(CustomIcon rii, int radio_index) {
    int pos_x = 0;
    int pos_y = 0;
    if(origin_x == 0) {
      pos_x = (networkpanel.getPreferredSize().width / 2);
      pos_y = (networkpanel.getPreferredSize().height / 2);
      origin_x = pos_x;
      origin_y = pos_y;
    } else {
      pos_x = origin_x;
      pos_y = origin_y;
    }
    int unit_x = rii.icon_get_width() + 20;
    int unit_y = rii.icon_get_height() + 20;
    int loop = 0;
    int loop_dec = 8;
    int loop_pos = 0;
    int loop_side = 0;
    int q1 = 0;
    int q2 = 0;
    int q3 = 0;
    int q4 = 0;
    while(radio_index > 0) {
      radio_index -= loop_dec;
      loop_pos = (loop_dec + radio_index) - 1;
      loop_dec += 8;
      loop++;
    }
    loop_side = (loop * 2);
    q1 = loop_side;
    q2 = (loop_side * 2);
    q3 = (loop_side * 3);
    q4 = (loop_side * 4);

    if(loop_pos <= q1) {
      pos_x -= (unit_x * loop);
      pos_y -= (unit_y * loop);

      pos_y += (unit_y * loop_pos);

    } else if(loop_pos <= q2) {
      pos_x -= (unit_x * loop);
      pos_y += (unit_y * loop);

      pos_x += (unit_x * (loop_pos - q1));
    } else if(loop_pos <= q3) {
      pos_x += (unit_x * loop);
      pos_y += (unit_y * loop);

      pos_y -= (unit_y * (loop_pos - q2));
    } else if(loop_pos <= q4) {
      pos_x += (unit_x * loop);
      pos_y -= (unit_y * loop);

      pos_x -= (unit_x * (loop_pos - q3));
    }
    pos_x = (pos_x / 10) * 10;
    pos_y = (pos_y / 10) * 10;

    rii.icon_set_x(pos_x);
    rii.icon_set_y(pos_y);
  }

  private CustomIcon radio_get_clicked(Point p) {
    CustomIcon ret = null;
    synchronized(mutex) {
      for(int i = 0;i < icons_zorder.size();i++) {
        CustomIcon ri = icons_zorder.get(i);
        double hs = peerlink_hit_size;
        Rectangle2D rmouse = new Rectangle2D.Double(p.x - (hs / 2), p.y - (hs / 2), hs, hs);
        Rectangle2D rradio = ri.icon_get_bounds();
        if(rradio.intersects(rmouse)) {
          ret = ri;
          break;
        }
      }
    }
    return ret;
  }

  //used by eHistory also
  public Iterator _radio_set_zorder(CustomIcon rii,int index) {
    int idx = icons_zorder.indexOf(rii);
    if(icons_zorder.get(index).equals(rii)){
      return null;
    }
    icons_zorder.remove(rii);
    icons_zorder.add(index, rii);
    return icons_zorder.listIterator(idx);
  }

  @SuppressWarnings("unchecked")
  // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
  private void initComponents() {

    scr_network = new javax.swing.JScrollPane();
    networkpanel = new javademo.CustomPanel(this);

    setDefaultCloseOperation(javax.swing.WindowConstants.EXIT_ON_CLOSE);
    setTitle("Java Demo");
    addFocusListener(new java.awt.event.FocusAdapter() {
      public void focusGained(java.awt.event.FocusEvent evt) {
        formFocusGained(evt);
      }
    });
    addComponentListener(new java.awt.event.ComponentAdapter() {
      public void componentResized(java.awt.event.ComponentEvent evt) {
        formComponentResized(evt);
      }
    });
    addWindowListener(new java.awt.event.WindowAdapter() {
      public void windowClosing(java.awt.event.WindowEvent evt) {
        formWindowClosing(evt);
      }
    });
    getContentPane().setLayout(null);

    scr_network.setBackground(java.awt.Color.white);
    scr_network.setHorizontalScrollBarPolicy(javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_ALWAYS);
    scr_network.setVerticalScrollBarPolicy(javax.swing.ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
    scr_network.setDoubleBuffered(true);
    scr_network.setMaximumSize(new java.awt.Dimension(32785, 32785));

    networkpanel.setBackground(new java.awt.Color(255, 255, 255));
    networkpanel.setPreferredSize(new java.awt.Dimension(32767, 32767));
    networkpanel.setLayout(null);
    scr_network.setViewportView(networkpanel);

    getContentPane().add(scr_network);
    scr_network.setBounds(0, 0, 791, 551);

    setBounds(0, 0, 800, 600);
  }// </editor-fold>//GEN-END:initComponents

  private void formComponentResized(java.awt.event.ComponentEvent evt) {//GEN-FIRST:event_formComponentResized
    Container c = this.getContentPane();
    Rectangle r = scr_network.getBounds();
    scr_network.setBounds(r.x, r.y,c.getWidth(), c.getHeight());
    networkpanel.revalidate();
    networkpanel.repaint();
  }//GEN-LAST:event_formComponentResized

  private void formWindowClosing(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosing
  }//GEN-LAST:event_formWindowClosing

  private void formFocusGained(java.awt.event.FocusEvent evt) {//GEN-FIRST:event_formFocusGained

  }//GEN-LAST:event_formFocusGained


  // Variables declaration - do not modify//GEN-BEGIN:variables
  public javax.swing.JPanel networkpanel;
  private javax.swing.JScrollPane scr_network;
  // End of variables declaration//GEN-END:variables

  @Override
  public void mouseMoved(MouseEvent e) {
  }

  @Override
  public void mouseClicked(MouseEvent e) {
  }

  @Override
  public void mouseEntered(MouseEvent e) {
  }

  @Override
  public void mouseExited(MouseEvent e) {
  }

  @Override
  public void keyTyped(KeyEvent e) {
  }

  @Override
  public void keyPressed(KeyEvent e) {
  }

  @Override
  public void keyReleased(KeyEvent e) {
  }

  @Override
  public void setVisible(boolean b) {
    super.setVisible(b);
    center_scroll();
  }
  
  /**
   * @param args the command line arguments
   */
  public static void main(String args[]) {
    /* Set the Nimbus look and feel */
        //<editor-fold defaultstate="collapsed" desc=" Look and feel setting code (optional) ">
        /* If Nimbus (introduced in Java SE 6) is not available, stay with the default look and feel.
     * For details see http://download.oracle.com/javase/tutorial/uiswing/lookandfeel/plaf.html
     */
    try {
      for (javax.swing.UIManager.LookAndFeelInfo info : javax.swing.UIManager.getInstalledLookAndFeels()) {
        if ("Nimbus".equals(info.getName())) {
          javax.swing.UIManager.setLookAndFeel(info.getClassName());
          break;
        }
      }
    } catch (ClassNotFoundException | InstantiationException | IllegalAccessException | javax.swing.UnsupportedLookAndFeelException ex) {
      java.util.logging.Logger.getLogger(MainFrame.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
    }
    //</editor-fold>
    //</editor-fold>
    //</editor-fold>
    //</editor-fold>

    /* Create and display the form */
    java.awt.EventQueue.invokeLater(new Runnable() {
      @Override
      public void run() {
        new MainFrame().setVisible(true);
      }
    });
  }
  
}
